"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from atomic_reactor import set_logging

set_logging(name="atomic_reactor.tests")
set_logging(name="atomic_reactor")
